#include "particle.h"

/*This class generates a very simple particle system.  It uses point sprites as each individual particle, then emits
the particles based on the time, and the rate at which the particles are supposed to be emitted.  Each particle has its
own class, which holds its specific information.  The particle system class manages all the particles, so they come 
out in a steady stream.

  Usage:
  *allocate
  *call CPARTICLESYSTEM::Init()
  call CPARTICLESYSTEM::Update()
  call CPARTICLESYSTEM::Draw()
  deallocate when done
 */


int intRandNum(int iMin, int iMax)
{
  if (iMin == iMax) return(iMin);
  return((rand() % (abs(iMax-iMin)+1))+iMin);
}

float floatRandNum(float fMin, float fMax)
{
  if (fMin == fMax) return(fMin);
  float fRandom = (float)rand() / (float)RAND_MAX;
  return((fRandom * (float)fabs(fMax-fMin))+fMin);
}

CPARTICLESYSTEM::CPARTICLESYSTEM()
{
	numActiveParts = 0;
	LeftOverParts = 0;
	Particles = new CPARTICLE[NUMPARTS];
}

CPARTICLESYSTEM::~CPARTICLESYSTEM()
{
	lpPartTex->Release();
	lpPartVertBuff->Release();
	delete[] Particles;
}

bool CPARTICLESYSTEM::Init(IDirect3DDevice8 *lpDevice, char * TexFilePath, int MinER, int MaxER, D3DXVECTOR3 Pos, float PartSpd, D3DXVECTOR3 vecPartDir1, D3DXVECTOR3 vecPartDir2)
{
	MinEmitRate = MinER;
	MaxEmitRate = MaxER;

	OriginPos = Pos;

	PartSpeed = PartSpd;

	PartDir1 = vecPartDir1;
	PartDir2 = vecPartDir2;

	HRESULT hRet;
	
	hRet = D3DXCreateTextureFromFile(lpDevice, TexFilePath, &lpPartTex);
	if(hRet != D3D_OK)
		return false;

	hRet = lpDevice->CreateVertexBuffer(NUMPARTS * sizeof(PARTVERT), D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY | D3DUSAGE_POINTS, FVF_PARTICLE, D3DPOOL_DEFAULT, &lpPartVertBuff);
	if(hRet != D3D_OK)
		return false;

	return true;
}

bool CPARTICLESYSTEM::Update(DWORD DeltaTime, bool bCreateNewParts)
{
	float fDeltaTime = (float)DeltaTime / 1000;
	int NewParts;

	for(int i = 0; i < NUMPARTS; i++)
	{
		if(Particles[i].IsActive == true)
		{
			if(!Particles[i].Update(fDeltaTime, PartSpeed))
			{
				Particles[i].IsActive = false;
				numActiveParts--;
			}
		}
	}

	float CurEmitRate = (float)intRandNum(MinEmitRate, MaxEmitRate);
	
	NewParts = (int)CurEmitRate * (int)fDeltaTime;

	LeftOverParts += (float)(CurEmitRate * fDeltaTime) - NewParts;

	if(LeftOverParts >= 1.0f)
	{
		NewParts += (int)LeftOverParts;
		LeftOverParts -= (int)LeftOverParts;
	}

	if(!bCreateNewParts)
		NewParts = 0;

	for(i = 0; i < NUMPARTS; i++)
	{
		if(NewParts <= 0)
		{
			i = NUMPARTS;
			break;	//break out of function since we have all particles we need
		}
		else
		{
			if(Particles[i].IsActive == false)
			{
				Particles[i].IsActive = true;

				float fRandX = floatRandNum(PartDir1.x, PartDir2.x);
			    float fRandY = floatRandNum(PartDir1.y, PartDir2.y);
			    float fRandZ = floatRandNum(PartDir1.z, PartDir2.z);

				Particles[i].Init(1.0f, 10.0f, OriginPos, D3DXVECTOR3(fRandX, fRandY, fRandZ), D3DCOLOR_RGBA(128, 128, 128, 128));

				NewParts--;
				numActiveParts++;
			}
		}
	}

	return true;
}

bool CPARTICLESYSTEM::Render(IDirect3DDevice8 *lpDevice)
{

	HRESULT hRet;

	float Psize = 100.0f, PSMin = 0.0f, PSMax = 1000.0f, PSA = 0.0f, PSB = 0.0f, PSC = 1.0f;
	DWORD OldSrcBlend, OldDestBlend, OldAlphaBlendEnable, OldZWriteEnable;

	lpDevice->SetRenderState( D3DRS_POINTSPRITEENABLE, TRUE );
	lpDevice->SetRenderState( D3DRS_POINTSCALEENABLE,  TRUE );
	lpDevice->SetRenderState( D3DRS_POINTSIZE,     *((DWORD*)&Psize));
    lpDevice->SetRenderState( D3DRS_POINTSIZE_MIN, *((DWORD*)&PSMin));
    lpDevice->SetRenderState( D3DRS_POINTSIZE_MAX, *((DWORD*)&PSMax));
    lpDevice->SetRenderState( D3DRS_POINTSCALE_A,  *((DWORD*)&PSA));
    lpDevice->SetRenderState( D3DRS_POINTSCALE_B,  *((DWORD*)&PSB));
    lpDevice->SetRenderState( D3DRS_POINTSCALE_C,  *((DWORD*)&PSC));

	lpDevice->SetStreamSource(0, lpPartVertBuff, sizeof(PARTVERT));
	lpDevice->SetVertexShader(FVF_PARTICLE);

	lpDevice->SetTexture(0, lpPartTex);

	PARTVERT *pVerts;

	if(numActiveParts > 0)
	{
	hRet = lpPartVertBuff->Lock(0, sizeof(PARTVERT) * numActiveParts, (BYTE**)&pVerts, D3DLOCK_DISCARD);
	if(hRet != D3D_OK)
		return false;

	int BuffPos = 0;
	for(int i = 0; i < NUMPARTS; i++)
	{
		if(Particles[i].IsActive == true)
		{
			pVerts[BuffPos].x = Particles[i].Pos.x;
			pVerts[BuffPos].y = Particles[i].Pos.y;
			pVerts[BuffPos].z = Particles[i].Pos.z;

			pVerts[BuffPos].Size = 1.0f;

			pVerts[BuffPos].dwColor = Particles[i].dwColor;//D3DCOLOR_RGBA(255, 255, 255, 255);
			BuffPos++;
		}
	}

	lpPartVertBuff->Unlock();
	}

	else
		return true;

	lpDevice->GetRenderState( D3DRS_ZWRITEENABLE, &OldZWriteEnable );
	lpDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE);

	lpDevice->GetRenderState( D3DRS_ALPHABLENDENABLE, &OldAlphaBlendEnable );
    lpDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );

	lpDevice->GetRenderState( D3DRS_SRCBLEND, &OldSrcBlend);
	lpDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);

	lpDevice->GetRenderState( D3DRS_DESTBLEND, &OldDestBlend);
    lpDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE);

	lpDevice->DrawPrimitive(D3DPT_POINTLIST, 0, numActiveParts);

	lpDevice->SetRenderState(D3DRS_ZWRITEENABLE, OldZWriteEnable );
    lpDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, OldAlphaBlendEnable);
	lpDevice->SetRenderState(D3DRS_SRCBLEND, OldSrcBlend);
    lpDevice->SetRenderState(D3DRS_DESTBLEND, OldDestBlend);

	lpDevice->SetRenderState( D3DRS_POINTSPRITEENABLE, FALSE );
	lpDevice->SetRenderState( D3DRS_POINTSCALEENABLE,  FALSE );

	return true;
}

CPARTICLE::CPARTICLE()
{
	IsActive = false;
}

CPARTICLE::~CPARTICLE()
{
}

inline bool CPARTICLE::Update(float fdTime, float PartSpeed)
{
	fAge += fdTime;

	if(fAge >= fLifeTime)
		return false;

	Alpha -= 100 * fdTime;
		
	if(Alpha < 0)
		Alpha = 0;

	BYTE bAlpha = (BYTE)Alpha;
	dwColor = D3DCOLOR_RGBA(255, 255, 255, bAlpha);

	Pos.x += (Dir.x * fdTime) * PartSpeed;
	Pos.y += (Dir.y * fdTime) * PartSpeed;
	Pos.z += (Dir.z * fdTime) * PartSpeed;

	return true;
}

void CPARTICLE::Init(float fPartSize, float fPartLifeTime, D3DXVECTOR3 vecPos, D3DXVECTOR3 vecDir, DWORD dwPartColor)
{
	fSize = fPartSize;
	fLifeTime = fPartLifeTime;
	Pos = vecPos;
	Dir = vecDir;
	dwColor = dwPartColor;
	fAge = 0;
	fLifeTime = 3;
	Alpha = 255;
}